/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015, 2016 (c)
 * 
 *    file: main.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include "defs.h"

const char tcalc_ver[] = "2.1";
int input_method;
#define INPUT_FROM_FILE	1
#define INPUT_FORMULA	2
#define INPUT_FREE	3

int number_mode = 0;
int decimal_point = 0;
int input_format = DECIMAL_FORMAT;
int output_format = DECIMAL_FORMAT;
int QUIET = 0;
int SCOPE = GLOBAL_SCOPE;

FILE *file;

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
	///////////////////////////////////////
	//parse command line arguments
	///////////////////////////////////////
	int c;
	char *file_name = NULL;
	char *formula = NULL;
	static int READ_FILE = 0, READ_FORMULA = 0;
	static int CONVERT = 0;
	char *num_to_convert = NULL;
	static int MULT_TABLE = 0;
	char *mult_num = NULL;
	static struct option long_options[] =
	{
	 {"input",   required_argument, &READ_FILE,  1},
	 {"version", no_argument,       0,         'v'},
	 {"help",    no_argument,       0,         'h'},
	 {"quiet",   no_argument,       0,         'q'},
	 {"decimal", required_argument, 0,         'd'},
	 {"format",  required_argument, 0,         'f'},
	 {"output",  required_argument, 0,         'o'},
	 {"convert", required_argument, 0,         'c'},
	 {"math",    no_argument,       0,         'm'},
	 {"table",   required_argument, 0,         't'},
	 {0, 0, 0, 0}
	};

	while(1)
	{
		int option_index=0;
		c = getopt_long(argc, argv, "i:vhqd:f:o:c:mt:", long_options, &option_index);
		if(c==-1) break;	//end of options

		switch(c)
		{
		case 0:
			if(long_options[option_index].flag != 0)
				if(strcmp(long_options[option_index].name, "input") == 0)
					file_name = optarg;
			break;
		case 'i':	//save file name so we can display info later
			file_name = optarg;
			READ_FILE = 1; break;
		case 'd':	//define decimal places
			decimal_point = atoi(optarg);
			break;
		case 't':	//print multiplication table
			mult_num = optarg;
			MULT_TABLE = 1;
			break;
		case 'c':	//convert a number
			CONVERT = 1;
			num_to_convert = optarg;
			break;
		case 'f':	//define input format
			if(strcmp(optarg, "d") == 0 || strcmp(optarg, "decimal") == 0)
			  input_format = DECIMAL_FORMAT;
			else if(strcmp(optarg, "x") == 0 || strcmp(optarg, "hex") == 0)
			  input_format = HEX_FORMAT;
			else if(strcmp(optarg, "b") == 0 || strcmp(optarg, "binary") == 0)
			  input_format = BINARY_FORMAT;
			else if(strcmp(optarg, "o") == 0 || strcmp(optarg, "octal") == 0)
			  input_format = OCTAL_FORMAT;
			else
			{
			  printf("Error: Unknown number format\n");
			  exit(1);
			}
			break;
		case 'o':	//define output format
			if(strcmp(optarg, "d") == 0 || strcmp(optarg, "decimal") == 0)
			  output_format = DECIMAL_FORMAT;
			else if(strcmp(optarg, "x") == 0 || strcmp(optarg, "hex") == 0)
			  output_format = HEX_FORMAT;
			else if(strcmp(optarg, "b") == 0 || strcmp(optarg, "binary") == 0)
			  output_format = BINARY_FORMAT;
			else if(strcmp(optarg, "o") == 0 || strcmp(optarg, "octal") == 0)
			  output_format = OCTAL_FORMAT;
			else if(strcmp(optarg, "a") == 0 || strcmp(optarg, "all") == 0)
			  output_format = ALL_FORMATS;
			else
			{
			  printf("Error: Unknown number format\n");
			  exit(1);
			}
			break;
		case 'v':	//show version & exit
			printf("%s\n", tcalc_ver);
			exit(0); break;
		case 'm':	//print math functions & exit
			print_func_descriptions();
			exit(0); break;
		case 'q':	//quiet output
			QUIET = 1;
			break;
		case 'h':	//show help & exit
			printf("Terminal Calculator for GNU/Linux console, version %s\n"
				"Usage: %s [options]\n\n"
				"Options:\n"
				"Formula              Formula to be calculated\n"
				"-c, --convert n      Convert n to the output format specified\n"
				"                     by '-o' below, or to decimal if none\n"
				"-d, --decimal n      How many digits are shown to the right\n"
				"                     of the decimal point\n"
				"-f, --format n       Input format, possible values are:\n"
				"                           * binary (b)\n"
				"                           * hex (x)\n"
				"                           * decimal (d)\n"
				"                           * octal (o)\n"
				"-h, --help           Print help [this page]\n"
				"-i, --input filename Name of file containing formulas\n"
				"-m, --math           Print a list of math functions\n"
				"-o, --output n       Output format, possible values are:\n"
				"                           * binary (b)\n"
				"                           * hex (x)\n"
				"                           * decimal (d)\n"
				"                           * octal (o)\n"
				"                           * all formats (a)\n"
				"-q, --quiet          Minimize output messages\n"
				"-t, --table n        Print the multiplication table of number 'n'\n"
				//"-t, --table [from] [to] [step] n\n"
				//"                     Print the multiplication table of number 'n', table ranges\n"
				//"                     between 'from' and 'to', with a stepping of 'step'\n"
				"-v, --version        Display program version\n",
				tcalc_ver, argv[0]);
			exit(0); break;
		case '?':
			break;
		default:
			abort();
		}
	}
	///////////////////////////////////////
	while(optind < argc)
	{
	  formula = argv[optind++];
	  READ_FORMULA = 1;
	}

	if(file_name == NULL)
	{
	  if(!READ_FORMULA)
	    input_method = INPUT_FREE;
	  else
	    input_method = INPUT_FORMULA;
	}
	else
	{
	  if((file = fopen(file_name, "rb")) == NULL)
	  {
	    fprintf(stderr, "Error opening file: %s\n", file_name);
	    exit(1);
	  }
	  else
	    input_method = INPUT_FROM_FILE;
	}
	
	if(CONVERT)
	{
	  convert_numbers(num_to_convert);
	  exit(0);
	}
	
	if(MULT_TABLE)
	{
	  /*********************
	   * TODO: Correct THIS!
	   *********************/
	  print_mult_table(mult_num, 1, 12, 1);
	  exit(0);
	}
	
	if(!QUIET)
	{
	  printf("Terminal Calculator for GNU/Linux console\n");
	  if(input_method == INPUT_FROM_FILE)
	    printf("Reading from %s\n", file_name);
	  else if(input_method == INPUT_FORMULA)
	    printf("Reading command line formula\n");
	  else
	    printf("Free input:\nEnter 'q' to quit, 'h' for help\n");
	}

	memset((void *)&custom_func_head, 0, sizeof(struct custom_func_s));
	total_custom_funcs = 0;
	total_global_vars = 0;
	///////////////////////////////////////
	///////////////////////////////////////
	if(input_method == INPUT_FROM_FILE)
	{
	  input_from_file_mode(file, 0);
	  fclose(file);
	}
	else if(input_method == INPUT_FREE)
	{
	  input_free_mode();
	}
	else if(input_method == INPUT_FORMULA)
	{
	  char COMMENT = 0;
	  char error = 0;
	  long result = 0;
	  double result_f = 0;
	  result_f = calc(formula, &error, &COMMENT, NULL);
	  if(number_mode == NUMBER_INT)
	    result = (long)result_f;
	  //////////////////////////////////////////////
	  //output the result
	  //////////////////////////////////////////////
	  if(error) exit(1);
          //printf("%s(): result = %ld (%lu), result_f = %g\n", __func__, result, result, result_f);
	  output_res(output_format, result_f, result);
	}//end INPUT_FORMULA
	
	return 0;
}


//////////////////////////////////////////////
// Pass the active function as a pointer to
// struct in 'active_func'. Pass NULL if we are
// in the global scope.
//////////////////////////////////////////////
char tmp3[1024];
double process_sub_formula_f(char *subformula, int end, char *error, struct custom_func_s *active_func)
{
  int i = 1, j, k;
  double res = 0;
  char op = '.';
  char *tmp = tmp3;
  *error = 0;
  for(k = 1; k <= end; k++)
  {
      if(subformula[k] == '(')
      {
	for(j = end-1; j > 0; j--)
	  if(subformula[j] == ')')
	  {
		double tmp_res = 
		  process_sub_formula_f(subformula+k, j-k, error, active_func);
		if(*error) return 0;
		__do_op_float(op, &res, tmp_res);
		break;
	  }
        if(j <= 0)
        {
            fprintf(stderr, "Error: Missing ')'\n");
            *error = 1;
            return 0;
        }
        op = subformula[j+1];
        k = j+1;
        i = j+2;
      }
      /////////////////////////////////
      /////////////////////////////////
      /////////////////////////////////
      else if(subformula[k] == '+' || subformula[k] == '-' ||
	      subformula[k] == '*' || subformula[k] == '/' ||
	      subformula[k] == '%' || subformula[k] == '\0'||
	      subformula[k] == ')' || subformula[k] == '&' ||
	      subformula[k] == '|' || subformula[k] == '\n')
      {
	//tmp = strndup(subformula+i, k-i);
	strncpy(tmp, subformula+i, k-i); tmp[k-i] = '\0';
	if(tmp == '\0') { *error = 2; return 0; }
	
	//check if this is a variable
	if(tmp[0] == 'a' || (tmp[0] > 'b' && tmp[0] <= 'z'))
	{
	  int m;
	  if(active_func)		/* function scope */
	  {
	    for(m = 0; m < active_func->total_vars; m++)
	    {
	      if(strcmp(active_func->var_name[m], tmp) == 0)
	      {
	        sprintf(tmp, "%f", active_func->var_value[m]); break;
	      }
	    }
	  }
	  /* not found. Try global */
	  if(!active_func || (active_func && m == active_func->total_vars))
	  {
	    for(m = 0; m < total_global_vars; m++)
	    {
	      if(strcmp(global_var_name[m], tmp) == 0)
	      {
	        sprintf(tmp, "%f", global_var_value[m]); break;
	      }
	    }
	    if(m == total_global_vars)
	    {
	      printf("Error: variable %s is not defined\n", tmp);
	      *error = 3;
	      return 0;
	    }
	  }
	}
	
	do_op(tmp, op, &res, error);
	op = subformula[k];
	i = k+1;
      }
  }//end for
  //printf("process_sub_formula_f(): res = %f\n", res);
  return res;
}


void print_mult_table(char *n, int from, int to, int step)
{
  //printf("In print_mult_table(), n = '%s'\n", n);
  long num;
  char ERROR;
  if(n[0] == 'b' || n[0] == 'B')
    num = binary_to_decimal(n+1, &ERROR);
  else if((n[0] == '0' && n[1] == 'X') ||
	  (n[0] == '0' && n[1] == 'x'))
    num = hex_to_decimal(n, &ERROR);
  else if(n[0] == '0' && n[1] != '.')
    num = octal_to_decimal(n, &ERROR);
  else num = atol(n);
  
  //printf("%ld, %d\n", num, number_mode);
  int i;
  char *r, *r2, *r3;
  for(i = from; i <= to; i += step)
  {
    if(output_format == DECIMAL_FORMAT || output_format == ALL_FORMATS)
      printf("%ld * %2d = %ld\n", num, i, num*i);
    else if(output_format == BINARY_FORMAT)
    {
      decimal_to_binary(num,   &r, &ERROR);
      decimal_to_binary(i,     &r2, &ERROR);
      decimal_to_binary(num*i, &r3, &ERROR);
      printf("b%s * b%s = b%s\n", r, r2, r3);
      free(r); free(r2); free(r3);
    }
    else if(output_format == HEX_FORMAT)
    {
      decimal_to_hex(num,   &r, &ERROR);
      decimal_to_hex(i,     &r2, &ERROR);
      decimal_to_hex(num*i, &r3, &ERROR);
      printf("%s * %s = %s\n", r, r2, r3);
      free(r); free(r2); free(r3);
    }
    else if(output_format == OCTAL_FORMAT)
      printf("0%o * 0%2o = 0%o\n", (int)num, (int)i, (int)(num*i));
  }
}

/*
 * Trim the given string, removing whitespace chars
 * from both ends. If KEEP_NEWLINE is TRUE, it will
 * keep newline char found at the end of the string.
 */
char *trim(char *n, char KEEP_NEWLINE)
{
  if(!n) return (char *)NULL;;
  //printf("Old '%s'(%d), ", n, strlen(n));
  int i = 0;
  int l = strlen(n);
  char NEWLINE = 0;
  if(!l) return (char *)NULL;
  if(KEEP_NEWLINE && n[l-1] == '\n') NEWLINE = 1;

  for(i = 0; i < l; i++)
  {
    if(n[i] == '\t' || n[i] == '\n' || n[i] == ' ') continue;
    else break;
  }
  n += i;
  l -= i;
  for(i = l-1; i >= 0; i--)
  {
    if(n[i] == '\t' || n[i] == '\n' || n[i] == ' ') n[i] = '\0';
    else break;
  }
  if(NEWLINE) { n[i+1] = '\n'; }
  //printf("New '%s'(%d)\n", n, strlen(n));
  return n;
}
